"""
This module provides functions for platform integration. Most of the functionality here
could also be achieved with psutils, but we want to avoid the large dependency.
"""

from __future__ import annotations

import logging
import os
import resource
import socket
import time
from pathlib import Path
from typing import Optional, Tuple, Union
from urllib.parse import urlparse

import requests

__all__ = [
    "cat",
    "get_inotify_limits",
    "CPU_CORE_COUNT",
    "cpu_usage_percent",
    "check_connection",
    "SystemdNotifier",
]


CPU_CORE_COUNT = os.cpu_count() or 1  # os.cpu_count can return None
LINUX_POWER_SUPPLY_PATH = "/sys/class/power_supply"


def cat(*paths: "os.PathLike[str]") -> Union[bytes, None]:
    """
    Attempts to read the content of multiple files which may not exist. Returns the
    content of the first file which can be read. If none of them can be read return
    None. Returns an integer if the content is a digit.
    """
    for path in paths:
        try:
            with open(path, "rb") as f:
                return f.read().strip()
        except OSError:
            pass

    return None


def get_inotify_limits() -> Tuple[int, int, int]:
    """
    Returns the current inotify limit settings as tuple.

    :returns: ``(max_user_watches, max_user_instances, max_queued_events)``
    :raises OSError: if the settings cannot be read from /proc/sys/fs/inotify. This may
        happen if /proc/sys is left out of the kernel image or simply not mounted.
    """
    root = Path("/proc/sys/fs/inotify")

    max_user_watches_path = root / "max_user_watches"
    max_user_instances_path = root / "max_user_instances"
    max_queued_events_path = root / "max_queued_events"

    max_user_watches = int(max_user_watches_path.read_bytes().strip())
    max_user_instances = int(max_user_instances_path.read_bytes().strip())
    max_queued_events = int(max_queued_events_path.read_bytes().strip())

    return max_user_watches, max_user_instances, max_queued_events


def cpu_usage_percent(interval: float = 0.1) -> float:
    """
    Returns a float representing the CPU utilization of the current process as a
    percentage. This duplicates the similar method from psutil to avoid the psutil
    dependency.

    Compares process times to system CPU times elapsed before and after the interval
    (blocking). It is recommended for accuracy that this function be called with an
    interval of at least 0.1 sec.

    A value > 100.0 can be returned in case of processes running multiple threads on
    different CPU cores. The returned value is explicitly NOT split evenly between all
    available logical CPUs. This means that a busy loop process running on a system with
    2 logical CPUs will be reported as having 100% CPU utilization instead of 50%.

    :param interval: Interval in sec between comparisons of CPU times.
    :returns: CPU usage during interval in percent.
    """
    if interval <= 0:
        raise ValueError(f"interval is not positive (got {interval!r})")

    def timer() -> float:
        return time.monotonic() * CPU_CORE_COUNT

    st1 = timer()
    rt1 = resource.getrusage(resource.RUSAGE_SELF)
    time.sleep(interval)
    st2 = timer()
    rt2 = resource.getrusage(resource.RUSAGE_SELF)

    delta_proc = (rt2.ru_utime - rt1.ru_utime) + (rt2.ru_stime - rt1.ru_stime)
    delta_time = st2 - st1

    try:
        overall_cpus_percent = (delta_proc / delta_time) * 100
    except ZeroDivisionError:
        return 0.0
    else:
        single_cpu_percent = overall_cpus_percent * CPU_CORE_COUNT
        return round(single_cpu_percent, 1)


def check_connection(
    hostname: str, timeout: int = 2, logger: Optional[logging.Logger] = None
) -> bool:
    """
    A low latency check for an internet connection.

    :param hostname: Hostname to use for connection check.
    :param timeout: Timeout in seconds for connection check.
    :param logger: If provided, log output for connection failures will be logged to
        this logger with the level DEBUG.
    :returns: Connection availability.
    """
    if urlparse(hostname).scheme not in ["http", "https"]:
        hostname = "http://" + hostname
    try:
        requests.head(hostname, timeout=timeout)
        return True
    except Exception:
        if logger:
            logger.debug("Could not reach %s", hostname, exc_info=True)
        return False


class SystemdNotifier:
    """
    An interface to notify the systemd the service manager about status changes.

    Sends a status message to the systemd the service manager on the socket address
    provided by the NOTIFY_SOCKET environment variable. Does nothing NOTIFY_SOCKET is
    not set.

    See https://www.freedesktop.org/software/systemd/man/sd_notify.html for a
    documentation of message formats expected by systemd.
    """

    def __init__(self) -> None:
        self._socket = None

        addr = os.getenv("NOTIFY_SOCKET")
        if addr is None:
            return
        elif addr[0] == "@":
            addr = "\0" + addr[1:]

        try:
            self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
            self._socket.connect(addr)
        except OSError:
            pass

    def notify(self, status: str) -> None:
        """
        Send a status update to the service manager.

        :param status: The status update to send.
        """
        if self._socket:
            try:
                self._socket.sendall(status.encode(errors="replace"))
            except OSError:
                pass



"""
Dataclasses for our internal and external APIs.
"""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import Enum

# ==== user ============================================================================


class AccountType(Enum):
    """Enum of account types"""

    Basic = "basic"
    Pro = "pro"
    Business = "business"
    Other = "other"


@dataclass
class Team:
    """A group of users with joint access to shared folders"""

    id: str
    """Unique identifier of the team"""
    name: str
    """Display name of the team"""


@dataclass
class RootInfo:
    """Namespace info for the root of a shared filesystem"""

    root_namespace_id: str
    """Unique ID of the user's root namespace"""
    home_namespace_id: str
    """Unique ID of the user's personal namespace

    This will be different from :attr:`root_namespace_id` when Maestral is set up to
    sync the shared folder of a team.
    """


@dataclass
class UserRootInfo(RootInfo):
    pass


@dataclass
class TeamRootInfo(RootInfo):
    home_path: str
    """Path of the user's personal home folder relative to the root namespace

    Only present for accounts set up as part of a team when syncing the entire team's
    folder.
    """


@dataclass
class Account:
    """Represents the user's account"""

    account_id: str
    """Unique account ID"""
    display_name: str
    """The user's name for display purposes"""
    email: str
    """The user's email address"""
    email_verified: bool
    """Whether the email address was verified"""
    profile_photo_url: str | None
    """A URL to the user's photo"""
    disabled: bool
    """Whether the account is disabled"""


@dataclass
class FullAccount(Account):
    """Represents the user's account and sync information"""

    country: str | None
    """The user's country"""
    locale: str
    """The user's locale"""
    team: Team | None
    """The team that a user belongs to, if any"""
    team_member_id: str | None
    """The member ID of user in a team, if any"""
    account_type: AccountType
    """The account type"""
    root_info: RootInfo
    """The user's root namespace to sync"""


@dataclass
class SpaceUsage:
    """Space usage information"""

    used: int
    """Space used by in bytes"""
    allocated: int
    """Space available in bytes"""


@dataclass
class PersonalSpaceUsage(SpaceUsage):
    """Space usage information for a user"""

    team_usage: SpaceUsage | None
    """Space usage of a user's team, if any"""


# ==== files ===========================================================================


class WriteMode(Enum):
    """Enum of write modes when uploading a file"""

    Add = "add"
    Update = "update"
    Overwrite = "overwrite"


@dataclass
class Metadata:
    """Base class for sync item metadata"""

    name: str
    """Name of the file or folder"""
    path_lower: str
    """Normalised path on the server"""
    path_display: str
    """Cased path for display purposes and the local file system"""


@dataclass
class DeletedMetadata(Metadata):
    """Metadata of a deleted item"""

    pass


@dataclass
class FileMetadata(Metadata):
    """File metadata"""

    id: str
    """Unique ID on the server"""
    client_modified: datetime
    """Modified time in UTC as provided by clients"""
    server_modified: datetime
    """Server-side modified time in UTC"""
    rev: str
    """Unique ID of this version of a file"""
    size: int
    """File size in bytes"""
    symlink_target: str | None
    """If the file is a symlink, path of the target relative to the root namespace"""
    shared: bool
    """Whether the file is shared"""
    modified_by: str | None
    """Unique ID of the account that created / modified this revision"""
    is_downloadable: bool
    """Whether the file can be downloaded"""
    content_hash: str
    """A content hash of the file"""


@dataclass
class FolderMetadata(Metadata):
    """Folder metadata"""

    id: str
    """Unique ID on the server"""
    shared: bool
    """Whether the folder is shared"""


@dataclass
class ListFolderResult:
    """Result from listing the contents of a folder"""

    entries: list[Metadata]
    """List of entries"""
    has_more: bool
    """Whether there are more entries than listed"""
    cursor: str
    """Cursor to iterate and fetch more entries"""


# ==== sharing =========================================================================


class LinkAccessLevel(Enum):
    """Enum of access levels to shared links"""

    Viewer = "viewer"
    Editor = "editor"
    Other = "other"


class LinkAudience(Enum):
    """Enum of shared link audience"""

    Public = "public"
    Team = "team"
    NoOne = "no_one"
    Other = "other"


@dataclass
class LinkPermissions:
    """Permissions for a shared link"""

    can_revoke: bool
    """If the link can be revoked"""
    allow_download: bool
    """If the link allows users to download the item"""
    effective_audience: LinkAudience
    """The effective audience of link (who can use it)"""
    link_access_level: LinkAccessLevel
    """The type of access that the link grants to the item (how they can use it)"""
    require_password: bool | None
    """Whether a password is required when accessing the item through this link

    Note that users who already have access to an item otherwise will not need a
    password regardless of this value."""


@dataclass
class SharedLinkMetadata:
    """Metadata for a shared link"""

    url: str
    """The URL string"""
    name: str
    """The basename of the item"""
    path_lower: str | None
    """The normalised path of the item"""
    expires: datetime | None
    """Expiry time for a link in UTC"""
    link_permissions: LinkPermissions
    """Permissions that a link grants its users"""


@dataclass
class ListSharedLinkResult:
    """Result from listing shared links"""

    entries: list[SharedLinkMetadata]
    """List of shared link metadata"""
    has_more: bool
    """Whether there are more items to fetch"""
    cursor: str
    """A cursor to continue iterating over shared links"""


# ==== update checks ===================================================================


@dataclass
class UpdateCheckResult:
    """Information on update availability"""

    update_available: bool
    """Whether an update to Maestral is available"""
    latest_release: str
    """The latest release that can be updated to"""
    release_notes: str
    """Release notes for all releases between the currently running version up to and
    including the latest version"""
